from enum import Enum

import litellm
import tiktoken
from dsp.modules.lm import LM as DSPyLM
from tenacity import Retrying, stop_after_attempt, wait_random

litellm.drop_params = True


class ModelName(str, Enum):
    GPT_3 = "gpt-3.5-turbo"
    GPT_4 = "gpt-4-turbo-preview"
    HAIKU = "claude-3-haiku-20240307"
    SONNET = "claude-3-sonnet-20240229"
    OPUS = "claude-3-opus-20240229"
    GEMINI = "gemini/gemini-pro"
    MISTRAL = "anyscale/mistralai/Mistral-7B-Instruct-v0.1"
    MIXTRAL = "anyscale/mistralai/Mixtral-8x7B-Instruct-v0.1"


MODEL = ModelName.GPT_3


def dspy_prompt(lm: DSPyLM) -> str:
    n = 1
    skip = 0
    provider: str = lm.provider
    last_prompt = None
    printed = []
    n = n + skip
    for x in reversed(lm.history[-100:]):
        prompt = x["prompt"]
        if prompt != last_prompt:
            if provider == "clarifai" or provider == "google":
                printed.append(
                    (
                        prompt,
                        x["response"],
                    ),
                )
            else:
                printed.append(
                    (
                        prompt,
                        x["response"].generations
                        if provider == "cohere"
                        else x["response"]["choices"],
                    ),
                )
        last_prompt = prompt
        if len(printed) >= n:
            break
    history_str = ""
    for idx, (prompt, choices) in enumerate(reversed(printed)):
        if (n - idx - 1) < skip:
            continue
        history_str += prompt
        text = ""
        if provider == "cohere":
            text = choices[0].text
        elif provider == "openai" or provider == "ollama":
            text = " " + lm._get_choice_text(choices[0]).strip()
        elif provider == "clarifai":
            text = choices
        elif provider == "google":
            text = choices[0].parts[0].text
        else:
            text = choices[0]["text"]
        history_str += text
        if len(choices) > 1:
            history_str += f" \t (and {len(choices)-1} other completions)"
    return history_str


def count_gpt_tokens(text: str, model: ModelName = MODEL) -> int:
    encoding = tiktoken.encoding_for_model(model)
    return len(encoding.encode(text))


def chat_message(role: str, content: str) -> dict[str, str]:
    return {"role": role, "content": content}


def system_message(content: str) -> dict[str, str]:
    return chat_message(role="system", content=content)


def user_message(content: str) -> dict[str, str]:
    return chat_message(role="user", content=content)


def assistant_message(content: str) -> dict[str, str]:
    return chat_message(role="assistant", content=content)


def oai_response(response) -> str:
    try:
        return response.choices[0].message.content
    except Exception:
        return response


def ai_retry_attempts(attempts: int = 3):
    return (
        Retrying(wait=wait_random(min=1, max=40), stop=stop_after_attempt(attempts))
        if attempts > 1
        else 1
    )



import json
from pathlib import Path
from typing import Callable
from uuid import uuid4

import chromadb
import torch
from chromadb import Collection as ChromaCollection
from chromadb.utils.embedding_functions import SentenceTransformerEmbeddingFunction
from langchain_core.documents import Document as LCDocument
from sentence_transformers import CrossEncoder
from termcolor import colored

CHROMA_EMBEDDING_MODEL = "multi-qa-mpnet-base-cos-v1"
CHROMA_DIR = "chroma_dir"
CHROMA_DEVICE = "cuda"
CHROMA_DELETE_EXISTING = False
CROSS_ENCODER_MODEL = "cross-encoder/ms-marco-MiniLM-L-6-v2"


def chroma_collection(
    name: str,
    persistent_dir: str = CHROMA_DIR,
    delete_existing: bool = CHROMA_DELETE_EXISTING,
    embedding_model: str = CHROMA_EMBEDDING_MODEL,
    device: str = CHROMA_DEVICE,
) -> ChromaCollection:
    chroma_client = chromadb.PersistentClient(path=persistent_dir)
    if delete_existing:
        try:
            chroma_client.delete_collection(name)
        except Exception as e:
            print(colored(f"Error deleting collection named {name}: {e}", "red"))
            pass
    embedding_function = SentenceTransformerEmbeddingFunction(
        model_name=embedding_model, device=device
    )
    collection = chroma_client.get_or_create_collection(
        name, embedding_function=embedding_function
    )
    return collection


def json_files_to_collection(
    flist: list[str | Path], chroma_collection: ChromaCollection
):
    ids = []
    docs = []
    metas = []
    for i, fname in enumerate(flist):
        with open(fname, "r") as f:
            data = json.load(f)
            ids.append(f"id_{uuid4()}")
            metas.append(
                {
                    "level": data["level"],
                    "type": data["type"],
                    "prev_id": "",
                    "next_id": "",
                }
            )
            docs.append(json.dumps(data))
    print(colored(f"Adding {i} files to collection", "cyan"))
    chroma_collection.add(ids=ids, documents=docs, metadatas=metas)


def id_from_lc_doc(doc: LCDocument, doc_idx=None, **kwargs) -> str | None:
    try:
        file_name = doc.metadata.get("filename", doc.metadata.get("source"))
        doc_id = file_name or ""
        page_number = doc.metadata.get("page_number")
        if page_number is not None:
            doc_id += f"_page_{page_number}"
        if doc_idx is not None:
            doc_id += f"_{doc_idx}"
        return doc_id + f"_{uuid4()}"
    except Exception as e:
        print(colored(f"Error getting id from LCDocument: {e}", "red"))
        return None


def lc_docs_to_chroma_docs(
    docs: list[LCDocument],
    id_fn: Callable = id_from_lc_doc,
    add_links: bool = True,
) -> tuple[list[str], list[str], list[dict]]:
    chroma_ids = []
    chroma_docs = []
    chroma_metadatas = []
    for i, doc in enumerate(docs):
        chroma_ids.append(id_fn(doc=doc, doc_idx=i) or f"id_{i}")
        chroma_docs.append(doc.page_content)
    for i, doc in enumerate(docs):
        metadata = {
            k: v for k, v in doc.metadata.items() if type(v) in [str, int, float, bool]
        }
        metadata["prev_id"] = ""
        metadata["next_id"] = ""
        if add_links:
            metadata["prev_id"] = chroma_ids[i - 1] if i > 0 else ""
            metadata["next_id"] = chroma_ids[i + 1] if i < len(docs) - 1 else ""
        chroma_metadatas.append(metadata)
    return {"ids": chroma_ids, "documents": chroma_docs, "metadatas": chroma_metadatas}


def traverse_id(
    id: str,
    collection: ChromaCollection,
    direction: str = "prev",
    n_steps: int = 2,
) -> list:
    ids = []
    for _ in range(n_steps):
        metadata = collection.get(ids=[id])["metadatas"][0]
        id = metadata.get(f"{direction}_id", "")
        if not id:
            break
        ids.append(id)
    return ids


def traverse_ids(
    ids: str | list[str],
    collection: ChromaCollection,
    n_prev_links: int = 2,
    n_next_links: int = 2,
) -> list[list[str]]:
    if isinstance(ids, str):
        ids = [ids]
    res_ids = []
    for id in ids:
        prev_ids = traverse_id(
            id=id,
            collection=collection,
            direction="prev",
            n_steps=n_prev_links,
        )
        next_ids = traverse_id(
            id=id,
            collection=collection,
            direction="next",
            n_steps=n_next_links,
        )
        res_ids.append(prev_ids[::-1] + [id] + next_ids)
    return res_ids


def rerank_chroma_results(
    query_text: str,
    results: dict,
    cross_encoder_model: str = CROSS_ENCODER_MODEL,
) -> dict:
    device = "cuda" if torch.cuda.is_available() else "cpu"
    cross_encoder = CrossEncoder(cross_encoder_model, device=device)
    pairs = [[query_text, doc] for doc in results["documents"][0]]
    scores = cross_encoder.predict(pairs)
    scores_idx = sorted(range(len(scores)), key=lambda i: scores[i], reverse=True)
    results = {k: [[v[0][i] for i in scores_idx]] for k, v in results.items() if v}
    return results


def query_collection(
    query_text: str,
    collection: ChromaCollection,
    n_results: int = 10,
    rerank_results: bool = False,
    n_prev_links: int = 2,
    n_next_links: int = 2,
    include: list[str] = ["metadatas", "documents"],
    reranker_model: str = CROSS_ENCODER_MODEL,
) -> tuple[list[dict], list[str]]:
    query_res = collection.query(
        query_texts=query_text, n_results=n_results, include=include
    )
    if rerank_results:
        query_res = rerank_chroma_results(
            query_text=query_text,
            results=query_res,
            cross_encoder_model=reranker_model,
        )
    init_ids = query_res["ids"][0]
    traversed_ids = traverse_ids(
        ids=init_ids,
        collection=collection,
        n_prev_links=n_prev_links,
        n_next_links=n_next_links,
    )
    results = [collection.get(ids=ids, include=include) for ids in traversed_ids]
    return results, init_ids



import inspect
import json
import re
import textwrap
import typing
from datetime import datetime
from functools import partial
from inspect import cleandoc
from itertools import chain
from pathlib import Path
from typing import Callable

import demoji
from IPython.display import Markdown
from unstructured.cleaners.core import (
    bytes_string_to_string,
    clean_non_ascii_chars,
    group_broken_paragraphs,
    group_bullet_paragraph,
    replace_mime_encodings,
    replace_unicode_quotes,
)


def flatten(o):
    "Concatenate all collections and items as a generator"
    for item in o:
        if isinstance(item, str):
            yield item
            continue
        try:
            yield from flatten(item)
        except TypeError:
            yield item


def extract_json(s):
    s = s[next(idx for idx, c in enumerate(s) if c in "{[") :]
    try:
        return json.loads(s)
    except json.JSONDecodeError as e:
        return json.loads(s[: e.pos])


def noop(x=None, *args, **kwargs):
    return x


def resolve_data_path(data_path: list | str | Path) -> chain:
    if not isinstance(data_path, list):
        data_path = [data_path]
    data_path = flatten(data_path)
    paths = []
    for dp in data_path:
        if isinstance(dp, (str, Path)):
            dp = Path(dp)
            if not dp.exists():
                raise Exception(f"Path {dp} does not exist.")
            if dp.is_dir():
                paths.append(dp.iterdir())
            else:
                paths.append([dp])
    return chain(*paths)


def flatten_list(my_list: list) -> list:
    """Flatten a list of lists."""
    new_list = []
    for x in my_list:
        if isinstance(x, list):
            new_list += flatten_list(x)
        else:
            new_list.append(x)
    return new_list


def deindent(text: str) -> str:
    return textwrap.dedent(cleandoc(text))


def remove_digits(text: str) -> str:
    return re.sub(r"\d+", "", text)


def clean_text(
    text: str, no_digits: bool = False, no_emojis: bool = False, group: bool = False
) -> str:
    text = re.sub(r"[\n+]", "\n", text)
    text = re.sub(r"[\t+]", " ", text)
    text = re.sub(r"[. .]", " ", text)
    text = re.sub(r"([ ]{2,})", " ", text)
    # print(text)
    try:
        text = bytes_string_to_string(text)
        # print(text)
    except Exception:
        pass
    try:
        text = clean_non_ascii_chars(text)
        # print(text)
    except Exception:
        pass
    try:
        text = replace_unicode_quotes(text)
        # print(text)
    except Exception:
        pass
    try:
        text = replace_mime_encodings(text)
        # print(text)
    except Exception:
        pass
    if group:
        try:
            text = "\n".join(group_bullet_paragraph(text))
            # print(text)
        except Exception:
            pass
        try:
            text = group_broken_paragraphs(text)
            # print(text)
        except Exception:
            pass
    if no_digits:
        text = remove_digits(text)
    if no_emojis:
        text = demoji.replace(text, "")
    return text.strip()


def get_param_names(func: Callable):
    func = func.func if type(func) == partial else func
    return inspect.signature(func).parameters.keys()


def get_required_param_names(func: Callable) -> list[str]:
    if type(func) == partial:
        params = inspect.signature(func.func).parameters
        return [
            name
            for name, param in params.items()
            if param.default == inspect.Parameter.empty
            and name not in func.keywords.keys()
        ]
    params = inspect.signature(func).parameters
    return [
        name
        for name, param in params.items()
        if param.default == inspect.Parameter.empty
    ]


def get_function_return_type(func: Callable) -> type:
    func = func.func if type(func) == partial else func
    sig = typing.get_type_hints(func)
    return sig.get("return", None)


def get_function_name(func: Callable) -> str:
    func = func.func if type(func) == partial else func
    return func.__name__


def get_function_source(func: Callable) -> str:
    func = func.func if type(func) == partial else func
    return inspect.getsource(func)


def get_function_info(func: Callable) -> str:
    """Get a string with the name, signature, and docstring of a function."""

    func = func.func if type(func) == partial else func
    name = func.__name__
    signature = inspect.signature(func)
    docstring = inspect.getdoc(func)
    desc = f"\n---\nA function:\nName: {name}\n"
    if signature:
        desc += f"Signature: {signature}\n"
    if docstring:
        desc += f"Docstring: {docstring}\n"
    return cleandoc(desc + "---\n\n")


def to_markdown(text: str) -> Markdown:
    return Markdown(
        textwrap.indent(text.replace("•", "  *"), "> ", predicate=lambda _: True)
    )


def current_time(format: str = "%m-%d-%Y_%H:%M:%S") -> str:
    # print(f"\n\nCURRENT TIME: {datetime.now().strftime(format)}\n\n")
    return datetime.now().strftime(format)


def sort_times(times, format="%m-%d-%Y_%H:%M:%S"):
    return sorted(times, key=lambda time: datetime.strptime(time, format))


def count_words(text: str) -> int:
    return len(text.split())


def count_lines(text: str) -> int:
    return len(text.split("\n"))


def token_count_to_word_count(token_count) -> int:
    return max(int(token_count * 0.75), 1)


def token_count_to_line_count(token_count) -> int:
    return max(int(token_count * 0.066), 1)


def word_count_to_token_count(word_count) -> int:
    return max(int(word_count / 0.75), 1)


def count_tokens(text: str) -> int:
    return word_count_to_token_count(len(text.split()))
